import pytest
from unittest.mock import patch

from hera.data import Op
from hera.utils import to_u16
from hera.vm import VirtualMachine


@pytest.fixture
def vm():
    return VirtualMachine()


def test_exec_one_delegates_to_lsl(vm):
    with patch("hera.vm.VirtualMachine.exec_lsl") as mock_exec_lsl:
        vm.exec_one(Op("LSL", ["R1", "R2"]))
        assert mock_exec_lsl.call_count == 1
        assert mock_exec_lsl.call_args == (("R1", "R2"), {})


def test_lsl_with_small_positive(vm):
    vm.registers[6] = 7
    vm.exec_lsl("R1", "R6")
    assert vm.registers[1] == 14


def test_lsl_with_large_positive(vm):
    vm.registers[6] = 15000
    vm.exec_lsl("R1", "R6")
    assert vm.registers[1] == 30000


def test_lsl_with_positive_overflow(vm):
    vm.registers[6] = 17000
    vm.exec_lsl("R1", "R6")
    assert vm.registers[1] == 34000


def test_lsl_with_small_negative(vm):
    vm.registers[6] = to_u16(-7)
    vm.exec_lsl("R1", "R6")
    assert vm.registers[1] == to_u16(-14)


def test_lsl_with_large_negative(vm):
    vm.registers[6] = to_u16(-8400)
    vm.exec_lsl("R1", "R6")
    assert vm.registers[1] == to_u16(-16800)


def test_lsl_with_negative_overflow(vm):
    vm.registers[6] = to_u16(-20000)
    vm.exec_lsl("R1", "R6")
    assert vm.registers[1] == 25536
    assert vm.flag_carry
    assert not vm.flag_overflow


def test_lsl_shifts_out_carry_when_blocked(vm):
    vm.registers[6] = to_u16(-20000)
    vm.flag_carry_block = True
    vm.exec_lsl("R1", "R6")
    assert vm.flag_carry


def test_lsl_shifts_in_carry(vm):
    vm.registers[6] = 7
    vm.flag_carry = True
    vm.exec_lsl("R1", "R6")
    assert vm.registers[1] == 15
    assert not vm.flag_carry


def test_lsl_ignores_carry_when_blocked(vm):
    vm.registers[6] = 7
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_lsl("R1", "R6")
    assert vm.registers[1] == 14
    assert not vm.flag_carry


def test_lsl_resets_carry(vm):
    vm.flag_carry = True
    vm.exec_lsl("R6", "R6")
    assert not vm.flag_carry


def test_lsl_does_not_affect_R0(vm):
    vm.registers[6] = 7
    vm.exec_lsl("R0", "R6")
    assert vm.registers[0] == 0


def test_lsl_increments_pc(vm):
    vm.exec_lsl("R6", "R6")
    assert vm.pc == 1


def test_lsl_sets_zero_flag(vm):
    vm.registers[6] = 32768
    vm.exec_lsl("R1", "R6")
    assert vm.registers[1] == 0
    assert vm.flag_zero
    assert not vm.flag_sign


def test_lsl_sets_sign_flag(vm):
    vm.registers[6] = 32767
    vm.exec_lsl("R1", "R6")
    assert vm.registers[1] == to_u16(-2)
    assert not vm.flag_zero
    assert vm.flag_sign


def test_lsl_ignores_overflow_flag(vm):
    vm.flag_overflow = True
    vm.exec_lsl("R1", "R6")
    assert vm.flag_overflow


def test_exec_one_delegates_to_lsr(vm):
    with patch("hera.vm.VirtualMachine.exec_lsr") as mock_exec_lsr:
        vm.exec_one(Op("LSR", ["R1", "R2"]))
        assert mock_exec_lsr.call_count == 1
        assert mock_exec_lsr.call_args == (("R1", "R2"), {})


def test_lsr_with_small_positive(vm):
    vm.registers[6] = 7
    vm.exec_lsr("R1", "R6")
    assert vm.registers[1] == 3


def test_lsr_with_large_positive(vm):
    vm.registers[6] = 15000
    vm.exec_lsr("R1", "R6")
    assert vm.registers[1] == 7500


def test_lsr_with_small_negative(vm):
    vm.registers[6] = to_u16(-7)
    vm.exec_lsr("R1", "R6")
    assert vm.registers[1] == 32764


def test_lsr_with_large_negative(vm):
    vm.registers[6] = to_u16(-8400)
    vm.exec_lsr("R1", "R6")
    assert vm.registers[1] == 28568


def test_lsr_with_another_large_nevative(vm):
    vm.registers[6] = to_u16(-20000)
    vm.exec_lsr("R1", "R6")
    assert vm.registers[1] == 22768


def test_lsr_shifts_out_carry_when_blocked(vm):
    vm.registers[6] = 3
    vm.flag_carry_block = True
    vm.exec_lsr("R1", "R6")
    assert vm.registers[1] == 1
    assert vm.flag_carry


def test_lsr_shifts_in_carry(vm):
    vm.registers[6] = 6
    vm.flag_carry = True
    vm.exec_lsr("R1", "R6")
    assert vm.registers[1] == to_u16(-32765)
    assert not vm.flag_carry


def test_lsr_ignores_carry_when_blocked(vm):
    vm.registers[6] = 6
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_lsr("R1", "R6")
    assert vm.registers[1] == 3
    assert not vm.flag_carry


def test_lsr_resets_carry(vm):
    vm.flag_carry = True
    vm.exec_lsr("R6", "R6")
    assert not vm.flag_carry


def test_lsr_does_not_affect_R0(vm):
    vm.registers[6] = 7
    vm.exec_lsr("R0", "R6")
    assert vm.registers[0] == 0


def test_lsr_increments_pc(vm):
    vm.exec_lsr("R6", "R6")
    assert vm.pc == 1


def test_lsr_sets_zero_flag(vm):
    vm.registers[6] = 1
    vm.exec_lsr("R1", "R6")
    assert vm.registers[1] == 0
    assert vm.flag_zero
    assert not vm.flag_sign


def test_lsr_sets_sign_flag(vm):
    vm.registers[6] = 6
    vm.flag_carry = True
    vm.exec_lsr("R1", "R6")
    assert not vm.flag_zero
    assert vm.flag_sign


def test_lsr_ignores_overflow_flag(vm):
    vm.flag_overflow = True
    vm.exec_lsr("R1", "R6")
    assert vm.flag_overflow


def test_exec_one_delegates_to_lsl8(vm):
    with patch("hera.vm.VirtualMachine.exec_lsl8") as mock_exec_lsl8:
        vm.exec_one(Op("LSL8", ["R1", "R2"]))
        assert mock_exec_lsl8.call_count == 1
        assert mock_exec_lsl8.call_args == (("R1", "R2"), {})


def test_lsl8_with_small_positive(vm):
    vm.registers[4] = 51
    vm.exec_lsl8("R3", "R4")
    assert vm.registers[3] == 13056


def test_lsl8_with_large_positive(vm):
    vm.registers[4] = 17000
    vm.exec_lsl8("R3", "R4")
    assert vm.registers[3] == 26624


def test_lsl8_with_small_negative(vm):
    vm.registers[4] = -4
    vm.exec_lsl8("R3", "R4")
    assert vm.registers[3] == to_u16(-1024)


def test_lsl8_with_large_negative(vm):
    vm.registers[4] = to_u16(-31781)
    vm.exec_lsl8("R3", "R4")
    assert vm.registers[3] == to_u16(-9472)


def test_lsl8_sets_zero_flag(vm):
    vm.registers[4] = 32768
    vm.exec_lsl8("R3", "R4")
    assert vm.registers[3] == 0
    assert vm.flag_zero
    assert not vm.flag_sign


def test_lsl8_sets_sign_flag(vm):
    vm.registers[4] = 32767
    vm.exec_lsl8("R3", "R4")
    assert vm.registers[3] == to_u16(-256)
    assert not vm.flag_zero
    assert vm.flag_sign


def test_lsl8_increments_pc(vm):
    vm.exec_lsl8("R1", "R1")
    assert vm.pc == 1


def test_lsl8_ignores_incoming_carry(vm):
    vm.flag_carry = True
    vm.registers[4] = 5
    vm.exec_lsl8("R3", "R4")
    assert vm.registers[3] == 1280
    assert vm.flag_carry


def test_lsl8_does_not_set_carry_or_overflow(vm):
    vm.registers[4] = to_u16(-1)
    vm.exec_lsl8("R3", "R4")
    assert vm.registers[3] == to_u16(-256)
    assert not vm.flag_carry
    assert not vm.flag_overflow


def test_lsl8_does_not_affect_R0(vm):
    vm.registers[4] = 4
    vm.exec_lsl8("R0", "R4")
    assert vm.registers[0] == 0


def test_exec_one_delegates_to_lsr8(vm):
    with patch("hera.vm.VirtualMachine.exec_lsr8") as mock_exec_lsr8:
        vm.exec_one(Op("LSR8", ["R1", "R2"]))
        assert mock_exec_lsr8.call_count == 1
        assert mock_exec_lsr8.call_args == (("R1", "R2"), {})


def test_lsr8_with_small_positive(vm):
    vm.registers[4] = 51
    vm.exec_lsr8("R3", "R4")
    assert vm.registers[3] == 0


def test_lsr8_with_large_positive(vm):
    vm.registers[4] = 17000
    vm.exec_lsr8("R3", "R4")
    assert vm.registers[3] == 66


def test_lsr8_with_small_negative(vm):
    vm.registers[4] = to_u16(-4)
    vm.exec_lsr8("R3", "R4")
    assert vm.registers[3] == 255


def test_lsr8_with_large_negative(vm):
    vm.registers[4] = to_u16(-31781)
    vm.exec_lsr8("R3", "R4")
    assert vm.registers[3] == 131


def test_lsr8_sets_zero_flag(vm):
    vm.registers[4] = 17
    vm.exec_lsr8("R3", "R4")
    assert vm.registers[3] == 0
    assert vm.flag_zero
    assert not vm.flag_sign


def test_lsr8_increments_pc(vm):
    vm.exec_lsr8("R1", "R1")
    assert vm.pc == 1


def test_lsr8_ignores_incoming_carry(vm):
    vm.flag_carry = True
    vm.registers[4] = 17000
    vm.exec_lsr8("R3", "R4")
    assert vm.registers[3] == 66
    assert vm.flag_carry


def test_lsr8_does_not_set_carry_or_overflow(vm):
    vm.registers[4] = 15910
    vm.exec_lsr8("R3", "R4")
    assert vm.registers[3] == 62
    assert not vm.flag_carry
    assert not vm.flag_overflow


def test_lsr8_does_not_affect_R0(vm):
    vm.registers[4] = 15910
    vm.exec_lsr8("R0", "R4")
    assert vm.registers[0] == 0


def test_exec_one_delegates_to_asl(vm):
    with patch("hera.vm.VirtualMachine.exec_asl") as mock_exec_asl:
        vm.exec_one(Op("ASL", ["R1", "R2"]))
        assert mock_exec_asl.call_count == 1
        assert mock_exec_asl.call_args == (("R1", "R2"), {})


def test_asl_with_small_positive(vm):
    vm.registers[6] = 7
    vm.exec_asl("R1", "R6")
    assert vm.registers[1] == 14


def test_asl_with_large_positive(vm):
    vm.registers[6] = 15000
    vm.exec_asl("R1", "R6")
    assert vm.registers[1] == 30000


def test_asl_with_positive_overflow(vm):
    vm.registers[6] = 17000
    vm.exec_asl("R1", "R6")
    assert vm.registers[1] == 34000


def test_asl_with_small_negative(vm):
    vm.registers[6] = to_u16(-7)
    vm.exec_asl("R1", "R6")
    assert vm.registers[1] == to_u16(-14)


def test_asl_with_large_negative(vm):
    vm.registers[6] = to_u16(-8400)
    vm.exec_asl("R1", "R6")
    assert vm.registers[1] == to_u16(-16800)


def test_asl_with_negative_overflow(vm):
    vm.registers[6] = to_u16(-20000)
    vm.exec_asl("R1", "R6")
    assert vm.registers[1] == 25536
    assert vm.flag_carry
    assert vm.flag_overflow


def test_asl_shifts_out_carry_when_blocked(vm):
    vm.registers[6] = to_u16(-20000)
    vm.flag_carry_block = True
    vm.exec_asl("R1", "R6")
    assert vm.flag_carry
    assert vm.flag_overflow


def test_asl_shifts_in_carry(vm):
    vm.registers[6] = 7
    vm.flag_carry = True
    vm.exec_asl("R1", "R6")
    assert vm.registers[1] == 15
    assert not vm.flag_carry


def test_asl_ignores_carry_when_blocked(vm):
    vm.registers[6] = 7
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_asl("R1", "R6")
    assert vm.registers[1] == 14
    assert not vm.flag_carry


def test_asl_resets_carry(vm):
    vm.flag_carry = True
    vm.exec_asl("R6", "R6")
    assert not vm.flag_carry


def test_asl_does_not_affect_R0(vm):
    vm.registers[6] = 7
    vm.exec_asl("R0", "R6")
    assert vm.registers[0] == 0


def test_asl_increments_pc(vm):
    vm.exec_asl("R6", "R6")
    assert vm.pc == 1


def test_asl_sets_zero_flag(vm):
    vm.registers[6] = 32768
    vm.exec_asl("R1", "R6")
    assert vm.registers[1] == 0
    assert vm.flag_zero
    assert not vm.flag_sign


def test_asl_sets_sign_flag(vm):
    vm.registers[6] = 32767
    vm.exec_asl("R1", "R6")
    assert vm.registers[1] == to_u16(-2)
    assert not vm.flag_zero
    assert vm.flag_sign


def test_asl_resets_overflow_flag(vm):
    vm.flag_overflow = True
    vm.exec_asl("R1", "R6")
    assert not vm.flag_overflow


def test_exec_one_delegates_to_asr(vm):
    with patch("hera.vm.VirtualMachine.exec_asr") as mock_exec_asr:
        vm.exec_one(Op("ASR", ["R1", "R2"]))
        assert mock_exec_asr.call_count == 1
        assert mock_exec_asr.call_args == (("R1", "R2"), {})


def test_asr_with_small_positive(vm):
    vm.registers[6] = 7
    vm.exec_asr("R1", "R6")
    assert vm.registers[1] == 3


def test_asr_with_large_positive(vm):
    vm.registers[6] = 15000
    vm.exec_asr("R1", "R6")
    assert vm.registers[1] == 7500


def test_asr_with_small_negative(vm):
    vm.registers[6] = to_u16(-7)
    vm.exec_asr("R1", "R6")
    assert vm.registers[1] == to_u16(-3)


def test_asr_with_another_small_negative(vm):
    vm.registers[6] = to_u16(-5)
    vm.exec_asr("R1", "R6")
    assert vm.registers[1] == to_u16(-2)


def test_asr_with_large_negative(vm):
    vm.registers[6] = to_u16(-8400)
    vm.exec_asr("R1", "R6")
    assert vm.registers[1] == to_u16(-4200)


def test_asr_shifts_out_carry_when_blocked(vm):
    vm.registers[6] = 3
    vm.flag_carry_block = True
    vm.exec_asr("R1", "R6")
    assert vm.flag_carry
    assert not vm.flag_overflow


def test_asr_ignores_incoming_carry(vm):
    vm.registers[6] = 4
    vm.flag_carry = True
    vm.exec_asr("R1", "R6")
    assert vm.registers[1] == 2
    assert not vm.flag_carry


def test_asr_ignores_carry_when_blocked(vm):
    vm.registers[6] = 4
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.exec_asr("R1", "R6")
    assert vm.registers[1] == 2
    assert not vm.flag_carry


def test_asr_resets_carry(vm):
    vm.flag_carry = True
    vm.exec_asr("R6", "R6")
    assert not vm.flag_carry


def test_asr_does_not_affect_R0(vm):
    vm.registers[6] = 7
    vm.exec_asr("R0", "R6")
    assert vm.registers[0] == 0


def test_asr_increments_pc(vm):
    vm.exec_asr("R6", "R6")
    assert vm.pc == 1


def test_asr_sets_zero_flag(vm):
    vm.registers[6] = 1
    vm.exec_asr("R1", "R6")
    assert vm.registers[1] == 0
    assert vm.flag_zero
    assert not vm.flag_sign


def test_asr_sets_sign_flag(vm):
    vm.registers[6] = to_u16(-20)
    vm.exec_asr("R1", "R6")
    assert vm.registers[1] == to_u16(-10)
    assert not vm.flag_zero
    assert vm.flag_sign


def test_asr_does_not_reset_overflow_flag(vm):
    vm.flag_overflow = True
    vm.exec_asr("R1", "R6")
    assert vm.flag_overflow



import pytest
from unittest.mock import patch

from hera.data import Op
from hera.parser import parse, parse_file, replace_escapes


def test_replace_escapes_with_one_escape():
    assert replace_escapes("abc\\ndef") == "abc\ndef"


def test_replace_escapes_with_many_escapes():
    assert replace_escapes('\\\\\\n\\ta\\"') == '\\\n\ta"'


def test_replace_escapes_with_invalid_escape():
    assert replace_escapes("\\c") == "\\c"


def test_parse_op():
    assert parse("SETLO(R1, 4)") == [Op("SETLO", ["R1", 4])]


def test_parse_op_with_label():
    parsed = parse("SETLO(R1, top)")
    assert parsed == [Op("SETLO", ["R1", "top"])]
    assert parsed[0].args[0].type == "REGISTER"
    assert parsed[0].args[1].type == "SYMBOL"


def test_parse_string():
    parsed = parse('LP_STRING("hello")')
    assert parsed == [Op("LP_STRING", ["hello"])]
    assert parsed[0].args[0].type == "STRING"


def test_parse_empty_string():
    assert parse('LP_STRING("")') == [Op("LP_STRING", [""])]


def test_parse_string_with_escapes():
    assert parse('LP_STRING("multiline\\nstring with quotes: \\"\\"")') == [
        Op("LP_STRING", ['multiline\nstring with quotes: ""'])
    ]


def test_parse_character_literal():
    assert parse("SETLO(R1, 'X')") == [Op("SETLO", ["R1", 88])]


def test_parse_character_literal_with_escape():
    assert parse("SETLO(R1, '\\t')") == [Op("SETLO", ["R1", 9])]


def test_parse_signed_number():
    assert parse("SETLO(R1, -12)") == [Op("SETLO", ["R1", -12])]


def test_parse_hex_number():
    assert parse("SETLO(R4, 0x5f)") == [Op("SETLO", ["R4", 0x5F])]


def test_parse_negative_hex_number():
    assert parse("SETLO(R7, -0x2B)") == [Op("SETLO", ["R7", -0x2B])]


def test_parse_binary_number():
    assert parse("SETLO(R5, 0b10110)") == [Op("SETLO", ["R5", 22])]


def test_parse_negative_binary_number():
    assert parse("SETLO(R5, -0b10110)") == [Op("SETLO", ["R5", -22])]


def test_parse_octal_number():
    assert parse("SETLO(R3, 0o173)") == [Op("SETLO", ["R3", 123])]


def test_parse_negative_octal_number():
    assert parse("SETLO(R3, -0o173)") == [Op("SETLO", ["R3", -123])]


def test_parse_octal_number_without_o():
    assert parse("SETLO(R3, 0173)") == [Op("SETLO", ["R3", 123])]


def test_parse_label_starting_with_register_name():
    assert parse("LABEL(R1_INIT)") == [Op("LABEL", ["R1_INIT"])]


def test_parse_single_line_comment():
    assert parse("SETLO(R1, 0)  // R1 <- 0") == [Op("SETLO", ["R1", 0])]


def test_parse_hera_boilerplate():
    assert parse(
        "#include <HERA.h>\nvoid HERA_main() {SETLO(R1, 42)}", expand_includes=False
    ) == [Op("#include", ["<HERA.h>"]), Op("SETLO", ["R1", 42])]


def test_parse_hera_boilerplate_weird_whitespace_and_spelling():
    assert parse(
        "#include <HERA.h>\nvoid   HeRA_mAin( \t)\n {\n\n}", expand_includes=False
    ) == [Op("#include", ["<HERA.h>"])]


def test_parse_hera_boilerplate_no_includes():
    assert parse("void HERA_main() {SETLO(R1, 42)}") == [Op("SETLO", ["R1", 42])]


def test_parse_hera_boilerplate_gives_warning():
    with patch("hera.utils._emit_msg") as mock_emit_warning:
        parse("void HERA_main() {SETLO(R1, 42)}")
        assert mock_emit_warning.call_count == 1
        assert "Warning" in mock_emit_warning.call_args[0][0]
        assert "HERA_main" in mock_emit_warning.call_args[0][0]
        assert "not necessary" in mock_emit_warning.call_args[0][0]


def test_parse_another_single_line_comments():
    program = """\
// Single-line comment
SETLO(R9, 42)
    """
    assert parse(program) == [Op("SETLO", ["R9", 42])]


def test_parse_multiline_comment():
    program = """\
/* Starts on this line
   ends on this one */
SETLO(R1, 1)
    """
    assert parse(program) == [Op("SETLO", ["R1", 1])]


def test_parse_include_amidst_instructions():
    program = 'SETLO(R1, 42)\n#include "whatever"\n'
    assert parse(program, expand_includes=False) == [
        Op("SETLO", ["R1", 42]),
        Op("#include", ['"whatever"']),
    ]


def test_parse_missing_comma():
    with pytest.raises(SystemExit):
        parse("ADD(R1, R2 R3)")


def test_parse_missing_parenthesis():
    with pytest.raises(SystemExit):
        parse("LSL8(R1, R1")


def test_parse_missing_end_quote():
    with pytest.raises(SystemExit):
        parse('LP_STRING("forgot to close my string)')


def test_parse_exception_has_line_number(capsys):
    program = "SETLO(R1, 10)\nSETHI(R1, 255)\nLSL(R1 R1)"
    with pytest.raises(SystemExit):
        parse(program)

    captured = capsys.readouterr().err
    assert "unexpected character" in captured
    assert "line 3 col 8 of <string>" in captured


def test_parse_expands_include():
    assert parse_file("test/assets/include/simple.hera", expand_includes=True) == [
        Op("BR", ["end_of_add"]),
        Op("LABEL", ["add"]),
        Op("ADD", ["R3", "R1", "R2"]),
        Op("RETURN", ["R12", "R13"]),
        Op("LABEL", ["end_of_add"]),
        Op("SET", ["R1", 20]),
        Op("SET", ["R2", 22]),
        Op("CALL", ["R12", "add"]),
    ]



import pytest

from hera.main import main
from hera.symtab import HERA_DATA_START
from hera.vm import VirtualMachine


def test_simple_include_program():
    vm = VirtualMachine()
    main(["test/assets/include/simple.hera"], vm)

    assert vm.registers[1] == 20
    assert vm.registers[2] == 22
    assert vm.registers[3] == 42

    for r in vm.registers[4:11]:
        assert r == 0

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_recursive_program(capsys):
    with pytest.raises(SystemExit):
        main(["test/assets/include/recursive.hera"])

    captured = capsys.readouterr()
    assert "recursive include" in captured.err
    assert '#include "recursive.hera"' in captured.err
    assert "line 1 col 10 of test/assets/include/recursive.hera" in captured.err


def test_mutually_recursive_programs(capsys):
    with pytest.raises(SystemExit):
        main(["test/assets/include/mutually_recursive1.hera"])

    captured = capsys.readouterr()
    assert "recursive include" in captured.err
    assert '#include "mutually_recursive1.hera"' in captured.err
    assert (
        "line 1 col 10 of test/assets/include/mutually_recursive2.hera" in captured.err
    )


def test_nonexistent_path_program(capsys):
    with pytest.raises(SystemExit):
        main(["test/assets/include/nonexistent_path.hera"])

    captured = capsys.readouterr()
    assert 'file "test/assets/include/whatever.hera" does not exist' in captured.err


def test_include_stdin_program(capsys):
    with pytest.raises(SystemExit):
        main(["test/assets/include/include_stdin.hera"])

    captured = capsys.readouterr()
    assert 'file "test/assets/include/-" does not exist' in captured.err


def test_include_before_data_program():
    vm = VirtualMachine()
    main(["test/assets/include/including_before_data.hera"], vm)

    assert vm.registers[1] == HERA_DATA_START
    assert vm.registers[2] == 42
    assert vm.registers[3] == 100

    for r in vm.registers[4:11]:
        assert r == 0

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block



import pytest
from io import StringIO
from unittest.mock import patch

from hera.data import Op
from hera.main import dump_state, main, main_preprocess, op_to_string, program_to_string
from hera.vm import VirtualMachine


def test_op_to_string():
    assert op_to_string(Op("SET", ["R1", "top"])) == "SET(R1, top)"


def test_op_to_string_with_integer():
    assert op_to_string(Op("INC", ["R7", 12])) == "INC(R7, 12)"


def test_program_to_string():
    program = [
        Op("SET", ["R1", 20]),
        Op("SET", ["R2", 22]),
        Op("ADD", ["R3", "R1", "R2"]),
    ]
    assert program_to_string(program) == "SET(R1, 20)\nSET(R2, 22)\nADD(R3, R1, R2)"


def test_dump_state(capsys):
    dump_state(VirtualMachine(), verbose=True)

    captured = capsys.readouterr()
    assert "R1  = 0x0000 = 0" in captured.err
    assert "R7  = 0x0000 = 0" in captured.err
    assert "R14 = 0x0000 = 0" in captured.err
    assert "Zero flag is OFF" in captured.err
    assert "Sign flag is OFF" in captured.err
    assert "Overflow flag is OFF" in captured.err
    assert "Carry flag is OFF" in captured.err
    assert "Carry-block flag is OFF" in captured.err


def test_main_non_existent_file(capsys):
    with pytest.raises(SystemExit):
        main("unicorn.hera")

    assert 'file "unicorn.hera" does not exist' in capsys.readouterr().err


def test_main_preprocess(capsys):
    # SET(R1, 10)
    main_preprocess("test/assets/unit/set.hera")

    assert capsys.readouterr().out == "SETLO(R1, 10)\nSETHI(R1, 0)\n"


def test_main_preprocess_non_existent_file(capsys):
    with pytest.raises(SystemExit):
        main_preprocess("unicorn.hera")

    assert 'file "unicorn.hera" does not exist' in capsys.readouterr().err


def test_execute_from_stdin():
    vm = VirtualMachine()

    with patch("sys.stdin", StringIO("SET(R1, 42)")):
        main(["-"], vm)

    assert vm.registers[1] == 42


def test_preprocess_from_stdin(capsys):
    with patch("sys.stdin", StringIO("SET(R1, 42)")):
        main(["preprocess", "-"])

    captured = capsys.readouterr()
    assert "SETLO(R1, 42)" in captured.out
    assert "SETHI(R1, 0)" in captured.out



import pytest
from unittest.mock import patch

from hera.debugger import Debugger
from hera.loader import load_program


@pytest.fixture
def debugger():
    return Debugger(SAMPLE_PROGRAM)


SAMPLE_PROGRAM = load_program("test/assets/unit/debugger.hera")


def test_print_breakpoints(debugger, capsys):
    debugger.breakpoints[4] = "main.hera:7"

    should_continue = debugger.handle_command("break")

    assert should_continue
    assert capsys.readouterr().out == "main.hera:7\n"


def test_print_breakpoints_with_no_breakpoints_set(debugger, capsys):
    should_continue = debugger.handle_command("break")

    assert should_continue
    assert capsys.readouterr().out == "No breakpoints set.\n"


def test_set_breakpoint(debugger):
    assert len(debugger.breakpoints) == 0

    should_continue = debugger.handle_command("break 2")

    assert should_continue
    assert len(debugger.breakpoints) == 1
    assert 0 in debugger.breakpoints
    assert debugger.breakpoints[0] == "test/assets/unit/debugger.hera:2"


def test_set_breakpoint_not_on_line_of_code(debugger, capsys):
    should_continue = debugger.handle_command("break 1")

    assert should_continue
    assert len(debugger.breakpoints) == 0
    assert capsys.readouterr().out == "Error: could not find corresponding line.\n"


def test_set_unparseable_breakpoint(debugger, capsys):
    should_continue = debugger.handle_command("break $$$")

    assert should_continue
    assert len(debugger.breakpoints) == 0
    assert capsys.readouterr().out == "Error: could not parse argument.\n"


def test_execute_break_with_too_many_args(debugger, capsys):
    should_continue = debugger.handle_command("break 1 2 3")

    assert should_continue
    assert len(debugger.breakpoints) == 0
    assert capsys.readouterr().out == "break takes zero or one arguments.\n"


def test_execute_abbreviated_break(debugger):
    with patch("hera.debugger.Debugger.exec_break") as mock_exec_break:
        debugger.handle_command("break 7")
        assert mock_exec_break.call_count == 1

        args = mock_exec_break.call_args[0]
        assert len(args) == 1
        assert args[0] == ["7"]

        kwargs = mock_exec_break.call_args[1]
        assert len(kwargs) == 0


def test_execute_next(debugger):
    assert debugger.vm.registers[1] == 0
    assert debugger.vm.pc == 0

    should_continue = debugger.handle_command("next")

    assert should_continue
    assert debugger.vm.registers[1] == 10
    assert debugger.vm.pc == 2


def test_execute_next_with_halt(debugger, capsys):
    # Last instruction of SAMPLE_PROGRAM is a HALT operation.
    debugger.vm.pc = len(debugger.program) - 1

    should_continue = debugger.handle_command("next")

    assert should_continue
    assert debugger.vm.pc == len(debugger.program) - 1
    assert capsys.readouterr().out == "Program has finished executing.\n"


def test_execute_next_after_end_of_program(debugger, capsys):
    debugger.vm.pc = 9000

    should_continue = debugger.handle_command("next")

    assert should_continue
    assert debugger.vm.pc == 9000
    assert (
        capsys.readouterr().out
        == "Program has finished executing. Press 'r' to restart.\n"
    )


def test_execute_next_with_too_many_args(debugger, capsys):
    # TODO: It would actually be useful if this worked.
    should_continue = debugger.handle_command("next 10")

    assert should_continue
    assert capsys.readouterr().out == "next takes no arguments.\n"


def test_execute_abbreviated_next(debugger):
    with patch("hera.debugger.Debugger.exec_next") as mock_exec_next:
        debugger.handle_command("n")
        assert mock_exec_next.call_count == 1


def test_execute_continue_with_breakpoint(debugger):
    debugger.breakpoints[4] = ""

    should_continue = debugger.handle_command("continue")

    assert should_continue
    assert debugger.vm.registers[1] == 10
    assert debugger.vm.registers[2] == 32
    assert debugger.vm.registers[3] == 0
    assert debugger.vm.pc == 4


def test_execute_continue_without_breakpoint(debugger, capsys):
    should_continue = debugger.handle_command("continue")

    assert should_continue
    assert debugger.vm.registers[1] == 10
    assert debugger.vm.registers[2] == 32
    assert debugger.vm.registers[3] == 42
    assert debugger.vm.pc == 5
    assert capsys.readouterr().out == "Program has finished executing.\n"


def test_execute_continue_with_too_many_args(debugger, capsys):
    should_continue = debugger.handle_command("continue 10")

    assert should_continue
    assert capsys.readouterr().out == "continue takes no arguments.\n"


def test_execute_abbreviated_continue(debugger):
    with patch("hera.debugger.Debugger.exec_continue") as mock_exec_continue:
        debugger.handle_command("c")
        assert mock_exec_continue.call_count == 1


def test_print_register(debugger, capsys):
    debugger.vm.registers[7] = 42

    should_continue = debugger.handle_command("print r7")

    assert should_continue
    assert capsys.readouterr().out == "r7 = 0x002a = 42 = '*'\n"


def test_print_invalid_register(debugger, capsys):
    should_continue = debugger.handle_command("print r17")

    assert should_continue
    assert capsys.readouterr().out == "r17 is not a valid register.\n"


def test_print_program_counter(debugger, capsys):
    debugger.vm.pc = 7

    should_continue = debugger.handle_command("print PC")

    assert should_continue
    assert capsys.readouterr().out == "PC = 7\n"


def test_print_memory_location(debugger, capsys):
    debugger.vm.assign_memory(97, 1000)

    should_continue = debugger.handle_command("print m[97]")

    assert should_continue
    assert capsys.readouterr().out == "M[97] = 1000\n"


def test_execute_print_with_too_few_args(debugger, capsys):
    # TODO: It would be useful if this printed the last printed expression again.
    should_continue = debugger.handle_command("print")

    assert should_continue
    assert capsys.readouterr().out == "print takes one argument.\n"


def test_execute_abbreviated_print(debugger):
    with patch("hera.debugger.Debugger.exec_print") as mock_exec_print:
        debugger.handle_command("p r7")
        assert mock_exec_print.call_count == 1

        args = mock_exec_print.call_args[0]
        assert len(args) == 1
        assert args[0] == ["r7"]

        kwargs = mock_exec_print.call_args[1]
        assert len(kwargs) == 0


def test_execute_quit(debugger):
    assert debugger.handle_command("quit") is False
    assert debugger.handle_command("q") is False


def test_execute_help(debugger, capsys):
    should_continue = debugger.handle_command("help")

    assert should_continue
    out = capsys.readouterr().out
    assert "Available commands" in out
    assert "Error:" not in out


def test_execute_unknown_command(debugger, capsys):
    should_continue = debugger.handle_command("whatever")

    assert should_continue
    assert capsys.readouterr().out == "whatever is not a known command.\n"


def test_resolve_location(debugger):
    assert debugger.resolve_location(2) == 0


def test_resolve_location_out_of_range(debugger):
    with pytest.raises(ValueError) as e:
        debugger.resolve_location(10)
    assert "could not find corresponding line" in str(e)


def test_resolve_location_invalid_format(debugger):
    with pytest.raises(ValueError) as e:
        debugger.resolve_location("a")
    assert "could not parse argument" in str(e)


def test_get_breakpoint_name(debugger):
    # Zero'th instruction corresponds to second line.
    assert debugger.get_breakpoint_name(0) == "test/assets/unit/debugger.hera:2"


def test_print_current_op(debugger, capsys):
    debugger.print_current_op()

    captured = capsys.readouterr()
    assert (
        captured.out
        == "[test/assets/unit/debugger.hera, line 2]\n\n0000  SET(R1, 10)\n"
    )



import pytest

from hera.main import main


def test_error_message_for_missing_comma(capsys):
    with pytest.raises(SystemExit):
        # SETLO(R1 40)
        main(["test/assets/error/missing_comma.hera"])

    captured = capsys.readouterr()
    assert "SETLO(R1 40)" in captured.err
    assert "line 1" in captured.err
    assert "col 10" in captured.err


def test_error_message_for_invalid_register(capsys):
    with pytest.raises(SystemExit):
        # SET(R17, 65)
        main(["test/assets/error/invalid_register.hera"])

    captured = capsys.readouterr()
    assert "SET(R17, 65)" in captured.err
    assert "line 1" in captured.err
    assert "col 5" in captured.err
    # Make sure the caret is aligned properly.
    assert "      ^" in captured.err
    assert "R17" in captured.err
    assert "not a valid register" in captured.err


def test_error_message_for_invalid_register_with_weird_syntax(capsys):
    with pytest.raises(SystemExit):
        # SET(
        # 	R17,
        # 	65)
        main(["test/assets/error/invalid_register_weird.hera"])

    captured = capsys.readouterr()
    assert "\tR17" in captured.err
    assert "SET(" not in captured.err
    assert "65" not in captured.err
    assert "line 2" in captured.err
    assert "col 2" in captured.err
    # Make sure the caret is aligned properly.
    assert "  \t^" in captured.err
    assert "not a valid register" in captured.err


def test_multiple_error_messages(capsys):
    with pytest.raises(SystemExit):
        # ADD(R1, 10)
        # INC(R4)
        main(["test/assets/error/multiple_errors.hera"])

    captured = capsys.readouterr()
    assert "ADD" in captured.err
    assert "too few" in captured.err
    assert "not a register" in captured.err
    assert "line 1" in captured.err
    assert "INC" in captured.err
    assert "line 2" in captured.err


def test_error_message_from_include(capsys):
    with pytest.raises(SystemExit):
        main(["test/assets/error/from_include.hera"])

    captured = capsys.readouterr()
    assert "SET(R1, R2)  // error!" in captured.err
    assert "ADD(R1, R2)  // error!" in captured.err
    assert "test/assets/error/from_include.hera" in captured.err
    assert "test/assets/error/included.hera" in captured.err


def test_dskip_overflow_program(capsys):
    with pytest.raises(SystemExit):
        main(["test/assets/error/dskip_overflow.hera"])

    captured = capsys.readouterr()
    assert "DSKIP(0xFFFF)" in captured.err
    assert "line 1" in captured.err


def test_warning_for_interrupt_instructions(capsys):
    main(["test/assets/error/interrupts.hera"])

    captured = capsys.readouterr()
    assert "SWI is a no-op in this simulator" in captured.err
    assert "line 2 col 1 of test/assets/error/interrupts.hera" in captured.err
    assert "SWI(10)" in captured.err
    assert "RTI is a no-op in this simulator" in captured.err
    assert "line 3 col 1 of test/assets/error/interrupts.hera" in captured.err
    assert "RTI()" in captured.err

    # Make sure that only one warning is given for each instruction.
    assert "line 6" not in captured.err
    assert "line 7" not in captured.err



"""Utilities for the HERA interpreter.

Author:  Ian Fisher (iafisher@protonmail.com)
Version: December 2018
"""
import os.path
import sys

from . import config
from .data import Token


def to_u16(n):
    """Reinterpret the signed integer `n` as a 16-bit unsigned integer.

    If `n` is too large for 16 bits, OverflowError is raised.
    """
    # Note that we allow positive values up to 2**16, but negative values only
    # down to -2**15.
    if n >= 2 ** 16 or n < -2 ** 15:
        raise OverflowError("signed integer too large for 16 bits")

    if n < 0:
        return 2 ** 16 + n
    else:
        return n


def from_u16(n):
    """Reinterpret the unsigned 16-bit integer `n` as a signed integer."""
    if n >= 2 ** 15:
        return -(2 ** 16 - n)
    else:
        return n


def to_u32(n):
    """Reinterpret the signed integer `n` as an unsigned 32-bit integer.

    If `n` is too large for 32 bits, OverflowError is raised.
    """
    # Note that we allow positive values up to 2**32, but negative values only
    # down to -2**31.
    if n >= 2 ** 32 or n < -2 ** 31:
        raise OverflowError("signed integer too large for 16 bits")

    if n < 0:
        return 2 ** 32 + n
    else:
        return n


NAMED_REGISTERS = {"rt": 11, "fp": 14, "sp": 15, "pc_ret": 13, "fp_alt": 12, "pc": -1}


def register_to_index(rname):
    """Return the index of the register with the given name in the register array."""
    original = rname
    rname = rname.lower()
    if rname in NAMED_REGISTERS:
        return NAMED_REGISTERS[rname]
    elif rname.startswith("r"):
        v = int(rname[1:])
        if 0 <= v < 16:
            return v
    raise ValueError("{} is not a valid register".format(original))


def is_register(s):
    return (s[0] in "rR" and s[1:].isdigit()) or s.lower() in NAMED_REGISTERS


def print_register_debug(target, v, *, to_stderr=True):
    file_ = sys.stderr if to_stderr else sys.stdout

    print("{0} = 0x{1:0>4x} = {1}".format(target, v), end="", file=file_)
    if v & 0x800:
        print(" = {}".format(from_u16(v)), end="", file=file_)
    if v < 128 and chr(v).isprintable():
        print(" = {!r}".format(chr(v)), end="", file=file_)
    print(file=file_)


def is_symbol(s):
    return isinstance(s, Token) and s.type == "SYMBOL"


REGISTER_BRANCHES = set(
    # Two lists instead of one so that the code formatter doesn't make this into a
    # dozen lines.
    ["BR", "BL", "BGE", "BLE", "BG", "BULE", "BUG", "BZ", "BNZ", "BC", "BNC", "BS"]
    + ["BNS", "BV", "BNV"]
)
RELATIVE_BRANCHES = set(b + "R" for b in REGISTER_BRANCHES)
DATA_STATEMENTS = set(["CONSTANT", "DLABEL", "INTEGER", "LP_STRING", "DSKIP"])


def emit_error(msg, *, loc=None, exit=False):
    """Print an error message to stderr."""
    msg = config.ANSI_RED_BOLD + "Error" + config.ANSI_RESET + ": " + msg
    config.ERROR_COUNT += 1
    _emit_msg(msg, loc=loc, exit=exit)


def emit_warning(msg, *, loc=None):
    """Print a error warning to stderr."""
    msg = config.ANSI_MAGENTA_BOLD + "Warning" + config.ANSI_RESET + ": " + msg
    config.WARNING_COUNT += 1
    _emit_msg(msg, loc=loc, exit=False)


def _emit_msg(msg, *, loc=None, exit=False):
    if loc is not None:
        if loc.path == "-":
            loc = loc._replace(path="<stdin>")

        if loc.line is not None:
            if loc.column is not None:
                caret = _align_caret(loc.file_lines[loc.line - 1], loc.column) + "^"
                msg += ", line {} col {} of {}\n\n  {}\n  {}\n".format(
                    loc.line, loc.column, loc.path, loc.file_lines[loc.line - 1], caret
                )
            else:
                msg += ", line {} of {}\n\n  {}\n".format(
                    loc.line, loc.path, loc.file_lines[loc.line - 1]
                )

    sys.stderr.write(msg + "\n")

    if exit:
        sys.exit(3)


def _align_caret(line, col):
    """Return the whitespace necessary to align a caret to underline the desired
    column in the line of text. Mainly this means handling tabs.
    """
    return "".join("\t" if c == "\t" else " " for c in line[: col - 1])


def get_canonical_path(fpath):
    if fpath == "-" or fpath == "<string>":
        return fpath
    else:
        return os.path.realpath(fpath)


def op_to_string(op):
    """Convert a single operation to a string."""
    return "{}({})".format(op.name, ", ".join(str(a) for a in op.args))



from unittest.mock import patch

from hera.data import IntToken, Op, Token
from hera.typechecker import (
    check_one_type,
    check_types,
    LABEL,
    REGISTER,
    REGISTER_OR_LABEL,
    STRING,
    typecheck,
    typecheck_one,
    U4,
    U16,
)


def R(s):
    return Token("REGISTER", s)


def SYM(s=""):
    return Token("SYMBOL", s)


def STR(s):
    return Token("STRING", s)


def test_check_types_with_too_few():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        check_types(SYM(), [REGISTER, REGISTER], [R("R1")], {})
        assert mock_emit_error.call_count == 1
        assert "too few" in mock_emit_error.call_args[0][0]


def test_check_types_with_too_many():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        check_types(SYM(), [REGISTER], [R("R1"), IntToken(10)], {})
        assert mock_emit_error.call_count == 1
        assert "too many" in mock_emit_error.call_args[0][0]


def test_check_types_with_wrong_type():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        check_types(SYM(), [REGISTER], [IntToken(10)], {})
        assert mock_emit_error.call_count == 1
        assert "not a register" in mock_emit_error.call_args[0][0]

    with patch("hera.utils._emit_msg") as mock_emit_error:
        check_types(SYM(), [U16], [R("R1")], {})
        assert mock_emit_error.call_count == 1
        assert "not an integer" in mock_emit_error.call_args[0][0]


def test_check_types_with_program_counter():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        check_types(SYM(), [REGISTER], [R("PC")], {})
        assert mock_emit_error.call_count == 1
        assert (
            "program counter cannot be accessed or changed directly"
            in mock_emit_error.call_args[0][0]
        )


def test_check_one_type_with_u16_out_of_range():
    assert check_one_type(U16, IntToken(65536), {}) == "out of range"


def test_check_one_type_with_negative_u16():
    assert check_one_type(U16, IntToken(-1), {}) == "must not be negative"


def test_check_one_type_with_u4_out_of_range():
    assert check_one_type(U4, IntToken(16), {}) == "out of range"
    assert check_one_type(U4, IntToken(-1), {}) == "must not be negative"


def test_check_one_type_with_range_object():
    assert check_one_type(range(-10, 10), IntToken(-11), {}) == "out of range"
    assert check_one_type(range(-10, 10), IntToken(10), {}) == "out of range"
    assert check_one_type(range(-10, 10), R("R1"), {}) == "not an integer"

    assert check_one_type(range(-10, 10), 5, {}) is None
    assert check_one_type(range(-10, 10), -10, {}) is None
    assert check_one_type(range(-10, 10), 9, {}) is None


def test_check_one_type_with_constant_symbol():
    assert check_one_type(range(0, 100), Token("SYMBOL", "n"), {"n": 42}) is None


def test_check_one_type_with_register_or_label():
    assert check_one_type(REGISTER_OR_LABEL, 5, {}) == "not a register or label"
    assert check_one_type(REGISTER_OR_LABEL, Token("SYMBOL", "n"), {"n": 42}) is None
    assert check_one_type(REGISTER_OR_LABEL, R("R1"), {}) is None


def test_check_one_type_with_invalid_register():
    assert check_one_type(REGISTER, R("R17"), {}) == "not a valid register"


def test_check_one_type_with_label():
    assert check_one_type(LABEL, 10, {}) == "not a symbol"
    assert check_one_type(LABEL, Token("SYMBOL", "n"), {"n": 42}) is None


def test_check_one_type_with_string():
    assert check_one_type(STRING, 10, {}) == "not a string"
    assert check_one_type(STRING, Token("STRING", "hello"), {}) is None


def test_typecheck_SET():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("SET", [R("R1"), 42]), {})
        typecheck_one(Op("SET", [R("R1"), 0xFFFF]), {})
        typecheck_one(Op("SET", [R("R1"), -0x7FFF]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_SETLO():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("SETLO", [R("R2"), 42]), {})
        typecheck_one(Op("SETLO", [R("R2"), 0xFF]), {})
        typecheck_one(Op("SETLO", [R("R2"), -0x7F]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_SETHI():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("SETHI", [R("R2"), 42]), {})
        typecheck_one(Op("SETHI", [R("R2"), 0xFF]), {})
        typecheck_one(Op("SETHI", [R("R2"), -0x7F]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_AND():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("AND", [R("R3"), R("R4"), R("R5")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_OR():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("OR", [R("R3"), R("R4"), R("R5")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_ADD():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("ADD", [R("R3"), R("R4"), R("R5")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_SUB():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("SUB", [R("R3"), R("R4"), R("R5")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_MUL():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("MUL", [R("R3"), R("R4"), R("R5")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_XOR():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("XOR", [R("R3"), R("R4"), R("R5")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_INC():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("INC", [R("R6"), 42]), {})
        typecheck_one(Op("INC", [R("R6"), 1]), {})
        typecheck_one(Op("INC", [R("R6"), 64]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_DEC():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("DEC", [R("R6"), 42]), {})
        typecheck_one(Op("DEC", [R("R6"), 1]), {})
        typecheck_one(Op("DEC", [R("R6"), 64]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_LSL():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("LSL", [R("R7"), R("R8")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_LSR():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("LSR", [R("R7"), R("R8")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_LSL8():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("LSL8", [R("R7"), R("R8")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_LSR8():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("LSR8", [R("R7"), R("R8")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_ASL():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("ASL", [R("R7"), R("R8")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_ASR():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("ASR", [R("R7"), R("R8")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_SAVEF():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("SAVEF", [R("R9")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_RSTRF():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("RSTRF", [R("R9")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_FON():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("FON", [0b10101]), {})
        typecheck_one(Op("FON", [0b01000]), {})
        typecheck_one(Op("FON", [0b11111]), {})
        typecheck_one(Op("FON", [0]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_FOFF():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("FOFF", [0b10101]), {})
        typecheck_one(Op("FOFF", [0b01000]), {})
        typecheck_one(Op("FOFF", [0b11111]), {})
        typecheck_one(Op("FOFF", [0]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_FSET5():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("FSET5", [0b10101]), {})
        typecheck_one(Op("FSET5", [0b01000]), {})
        typecheck_one(Op("FSET5", [0b11111]), {})
        typecheck_one(Op("FSET5", [0]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_FSET4():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("FSET4", [0b1010]), {})
        typecheck_one(Op("FSET4", [0b0100]), {})
        typecheck_one(Op("FSET4", [0b1111]), {})
        typecheck_one(Op("FSET4", [0]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_LOAD():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("LOAD", [R("R1"), 0, R("R2")]), {})
        typecheck_one(Op("LOAD", [R("R1"), 0b11111, R("R2")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_STORE():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("STORE", [R("R1"), 0, R("R2")]), {})
        typecheck_one(Op("STORE", [R("R1"), 0b11111, R("R2")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_CALL():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("CALL", [R("R12"), R("R11")]), {})
        typecheck_one(Op("CALL", [R("R12"), SYM("f")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_RETURN():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("RETURN", [R("R12"), R("R11")]), {})
        typecheck_one(Op("RETURN", [R("R12"), SYM("f")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BR():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BR", [R("R11")]), {})
        typecheck_one(Op("BR", [SYM("l")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BRR():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BRR", [0xFF]), {})
        typecheck_one(Op("BRR", [-0x7F]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BL():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BL", [R("R11")]), {})
        typecheck_one(Op("BL", [SYM("l")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BLR():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BLR", [0xFF]), {})
        typecheck_one(Op("BLR", [-0x7F]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BGE():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BGE", [R("R11")]), {})
        typecheck_one(Op("BGE", [SYM("l")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BGER():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BGER", [0xFF]), {})
        typecheck_one(Op("BGER", [-0x7F]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BLE():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BLE", [R("R11")]), {})
        typecheck_one(Op("BLE", [SYM("l")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BLER():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BLER", [0xFF]), {})
        typecheck_one(Op("BLER", [-0x7F]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BG():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BG", [R("R11")]), {})
        typecheck_one(Op("BG", [SYM("l")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BGR():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BGR", [0xFF]), {})
        typecheck_one(Op("BGR", [-0x7F]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BULE():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BULE", [R("R11")]), {})
        typecheck_one(Op("BULE", [SYM("l")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BULER():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BULER", [0xFF]), {})
        typecheck_one(Op("BULER", [-0x7F]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BUG():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BUG", [R("R11")]), {})
        typecheck_one(Op("BUG", [SYM("l")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BUGR():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BUGR", [0xFF]), {})
        typecheck_one(Op("BUGR", [-0x7F]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BZ():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BZ", [R("R11")]), {})
        typecheck_one(Op("BZ", [SYM("l")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BZR():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BZR", [0xFF]), {})
        typecheck_one(Op("BZR", [-0x7F]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BNZ():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BNZ", [R("R11")]), {})
        typecheck_one(Op("BNZ", [SYM("l")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BNZR():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BNZR", [0xFF]), {})
        typecheck_one(Op("BNZR", [-0x7F]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BC():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BC", [R("R11")]), {})
        typecheck_one(Op("BC", [SYM("l")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BCR():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BCR", [0xFF]), {})
        typecheck_one(Op("BCR", [-0x7F]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BNC():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BNC", [R("R11")]), {})
        typecheck_one(Op("BNC", [SYM("l")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BNCR():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BNCR", [0xFF]), {})
        typecheck_one(Op("BNCR", [-0x7F]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BS():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BS", [R("R11")]), {})
        typecheck_one(Op("BS", [SYM("l")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BSR():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BSR", [0xFF]), {})
        typecheck_one(Op("BSR", [-0x7F]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BNS():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BNS", [R("R11")]), {})
        typecheck_one(Op("BNS", [SYM("l")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BNSR():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BNSR", [0xFF]), {})
        typecheck_one(Op("BNSR", [-0x7F]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BV():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BV", [R("R11")]), {})
        typecheck_one(Op("BV", [SYM("l")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BVR():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BVR", [0xFF]), {})
        typecheck_one(Op("BVR", [-0x7F]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BNV():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BNV", [R("R11")]), {})
        typecheck_one(Op("BNV", [SYM("l")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_BNVR():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("BNVR", [0xFF]), {})
        typecheck_one(Op("BNVR", [-0x7F]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_SETRF():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("SETRF", [R("R1"), 42]), {})
        typecheck_one(Op("SETRF", [R("R1"), 0xFFFF]), {})
        typecheck_one(Op("SETRF", [R("R1"), -0x7FFF]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_MOVE():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("MOVE", [R("R1"), R("R2")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_CMP():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("CMP", [R("R1"), R("R2")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_NEG():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("NEG", [R("R1"), R("R2")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_NOT():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("NOT", [R("R1"), R("R2")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_CBON():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("CBON", []), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_CON():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("CON", []), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_COFF():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("COFF", []), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_CCBOFF():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("CCBOFF", []), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_FLAGS():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("FLAGS", [R("R1")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_NOP():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("NOP", []), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_HALT():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("HALT", []), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_LABEL():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("LABEL", [SYM("l")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_CONSTANT():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("CONSTANT", [SYM("N"), 0xFFFF]), {})
        typecheck_one(Op("CONSTANT", [SYM("N"), -0x7FFF]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_DLABEL():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("DLABEL", [SYM("l")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_INTEGER():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("INTEGER", [0xFFFF]), {})
        typecheck_one(Op("INTEGER", [-0x7FFF]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_LP_STRING():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("LP_STRING", [Token("STRING", "hello!")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_DSKIP():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("DSKIP", [0xFFFF]), {})
        typecheck_one(Op("DSKIP", [0]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_SWI():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("SWI", [0b1111]), {})
        typecheck_one(Op("SWI", [0b0110]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_RTI():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("RTI", []), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_print_reg():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op("print_reg", [R("R1")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_print():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op(SYM("print"), [STR("hello, world!")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_println():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op(SYM("println"), [STR("hello, world!")]), {})
        assert mock_emit_error.call_count == 0


def test_typecheck_undefined_symbol():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op(SYM("SET"), [R("R1"), SYM("N")]), {})
        assert mock_emit_error.call_count == 1
        assert "undefined constant" in mock_emit_error.call_args[0][0]


def test_typecheck_unknown_instruction():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op(SYM("IF"), [R("R1")]), {})
        assert mock_emit_error.call_count == 1
        assert "unknown instruction" in mock_emit_error.call_args[0][0]
        assert "IF" in mock_emit_error.call_args[0][0]


def test_typecheck_unknown_branch_instruction():
    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck_one(Op(SYM("BNWR"), [R("R1")]), {})
        assert mock_emit_error.call_count == 1
        assert "unknown instruction" in mock_emit_error.call_args[0][0]
        assert "BNWR" in mock_emit_error.call_args[0][0]


def test_typecheck_single_error():
    # Second argument to SETHI is out of range.
    program = [
        Op(SYM("SETLO"), [R("R1"), IntToken(10)]),
        Op(SYM("SETHI"), [R("R1"), IntToken(1000)]),
    ]

    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck(program, {})
        assert mock_emit_error.call_count == 1
        assert "SETHI" in mock_emit_error.call_args[0][0]
        assert "out of range" in mock_emit_error.call_args[0][0]


def test_typecheck_multiple_errors():
    program = [Op(SYM("ADD"), [R("R1"), IntToken(10)]), Op(SYM("INC"), [R("R3")])]

    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck(program, {})
        assert mock_emit_error.call_count == 3

        call_args = mock_emit_error.call_args_list[0][0]
        assert "ADD" in call_args[0]
        assert "too few" in call_args[0]

        call_args = mock_emit_error.call_args_list[1][0]
        assert "ADD" in call_args[0]
        assert "not a register" in call_args[0]

        call_args = mock_emit_error.call_args_list[2][0]
        assert "INC" in call_args[0]
        assert "too few" in call_args[0]


def test_typecheck_data_statement_after_instruction():
    program = [Op("SET", [R("R1"), 42]), Op(SYM("DLABEL"), [SYM("N")])]

    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck(program, {})
        assert mock_emit_error.call_count == 1
        assert "data statement after instruction" in mock_emit_error.call_args[0][0]


def test_typecheck_relative_branch_with_label():
    program = [Op(SYM("BRR"), [SYM("l")])]

    with patch("hera.utils._emit_msg") as mock_emit_error:
        typecheck(program, {"l": 7})
        assert mock_emit_error.call_count == 1
        assert "relative branches cannot use labels" in mock_emit_error.call_args[0][0]
        assert "why not use BR instead" in mock_emit_error.call_args[0][0]



import pytest

from hera.data import Op
from hera.vm import VirtualMachine


@pytest.fixture
def vm():
    return VirtualMachine()


def test_execute_unknown_instruction(vm):
    with pytest.raises(RuntimeError):
        vm.exec_one(Op("whatever", []))



"""hera: an interpreter for the Haverford Educational RISC Architecture.

Usage:
    hera [--lines=<n>] [-q | --verbose] [--no-color] <path>
    hera [--no-color] preprocess <path>
    hera [--no-color] debug <path>
    hera (-h | --help)
    hera (-v | --version)

Options:
    --lines=<n>      Only execute the first n lines of the program.
    --verbose        Set output level to verbose.
    -q --quiet       Set output level to quiet.
    --no-color       Do not print colored output.
    -h, --help       Show this message.
    -v, --version    Show the version.
"""
import sys
import functools

from docopt import docopt

from . import config
from .debugger import debug
from .loader import load_program
from .utils import op_to_string, print_register_debug
from .vm import VirtualMachine


LINES = None


def main(argv=None, vm=None):
    """The main entry point into hera-py.

    This function consists mostly of argument parsing. The heavy-lifting begins
    with main_execute later in this module.

    A virtual machine may be passed in for testing purposes. Otherwise, a new virtual
    machine is created.
    """
    arguments = docopt(__doc__, argv=argv, version="hera-py 0.4.0 for HERA version 2.4")
    path = arguments["<path>"]

    if arguments["--no-color"]:
        config.ANSI_MAGENTA_BOLD = config.ANSI_RED_BOLD = config.ANSI_RESET = ""

    if arguments["preprocess"]:
        main_preprocess(path)
    elif arguments["debug"]:
        main_debug(path)
    else:
        lines_to_exec = int(arguments["--lines"]) if arguments["--lines"] else None
        main_execute(
            path,
            lines_to_exec=lines_to_exec,
            verbose=arguments["--verbose"],
            quiet=arguments["--quiet"],
            vm=vm,
        )


def main_debug(path):
    """Debug the program."""
    program = load_program(path)
    debug(program)


def main_execute(path, *, lines_to_exec=None, verbose=False, quiet=False, vm=None):
    """Execute the program with the given options, most of which correspond to
    command-line arguments.

    A virtual machine instance may be passed in for testing purposes. If it is not, a
    new one is instantiated. The virtual machine is returned.
    """
    if vm is None:
        vm = VirtualMachine()

    program = load_program(path)

    vm.exec_many(program, lines=lines_to_exec)

    if not quiet:
        dump_state(vm, verbose=verbose)

    return vm


def main_preprocess(path):
    """Preprocess the program and print it to standard output."""
    program = load_program(path)
    print(program_to_string(program))


def program_to_string(ops):
    """Convert the list of operations to a string."""
    return "\n".join(op_to_string(op) for op in ops)


def dump_state(vm, *, verbose=False):
    """Print the state of the virtual machine to standard output."""
    sys.stdout.flush()

    # Redefine print in this function to use stderr.
    nprint = functools.partial(print, file=sys.stderr)

    if verbose:
        last_register = 15
    else:
        last_register = 10
        while last_register > 0 and vm.registers[last_register] == 0:
            last_register -= 1

    nprint("\nVirtual machine state after execution:")
    for i, value in enumerate(vm.registers[1 : last_register + 1], start=1):
        rname = "\tR" + str(i) + (" " if i < 10 else "")
        print_register_debug(rname, value)

    nprint()

    flags = [
        vm.flag_carry_block,
        vm.flag_carry,
        vm.flag_overflow,
        vm.flag_zero,
        vm.flag_sign,
    ]
    if not verbose and all(flags):
        nprint("\tAll flags are ON")
    elif not verbose and all(not f for f in flags):
        nprint("\tAll flags are OFF")
    else:
        nprint("\tCarry-block flag is " + ("ON" if vm.flag_carry_block else "OFF"))
        nprint("\tCarry flag is " + ("ON" if vm.flag_carry else "OFF"))
        nprint("\tOverflow flag is " + ("ON" if vm.flag_overflow else "OFF"))
        nprint("\tZero flag is " + ("ON" if vm.flag_zero else "OFF"))
        nprint("\tSign flag is " + ("ON" if vm.flag_sign else "OFF"))

    if config.WARNING_COUNT > 0:
        c = config.WARNING_COUNT
        nprint("\n{} warning{} emitted.".format(c, "" if c == 1 else "s"))



from .preprocessor import convert
from .utils import emit_error


# Arbitrary value copied over from HERA-C.
HERA_DATA_START = 0xC001


def get_symtab(program):
    """Return the program's symbol table, a dictionary mapping from strings to Label,
    DataLabel and Constant objects (all subclasses of int).
    """
    symtab = {}
    pc = 0
    dc = HERA_DATA_START
    for op in program:
        odc = dc
        if op.name == "LABEL":
            if len(op.args) == 1:
                update_labels(symtab, op.args[0], Label(pc), op)
        elif op.name == "DLABEL":
            if len(op.args) == 1:
                update_labels(symtab, op.args[0], DataLabel(dc), op)
        elif op.name == "CONSTANT":
            if len(op.args) == 2:
                update_labels(symtab, op.args[0], Constant(op.args[1]), op)
        elif op.name == "INTEGER":
            dc += 1
        elif op.name == "DSKIP":
            if len(op.args) == 1:
                if isinstance(op.args[0], int):
                    dc += op.args[0]
                elif op.args[0] in symtab and isinstance(symtab[op.args[0]], Constant):
                    dc += symtab[op.args[0]]
        elif op.name == "LP_STRING":
            if len(op.args) == 1 and isinstance(op.args[0], str):
                dc += len(op.args[0]) + 1
        else:
            pc += len(convert(op))

        if dc >= 0xFFFF and odc < 0xFFFF:
            emit_error("past the end of available memory", loc=op.name.location)

    return symtab


def update_labels(labels, k, v, op):
    if k in labels:
        emit_error(
            "symbol `{}` has already been defined".format(k), loc=op.name.location
        )
    else:
        labels[k] = v


class Label(int):
    pass


class DataLabel(int):
    pass


class Constant(int):
    pass



"""Debug HERA programs.

`debug` is the sole public function.

Author:  Ian Fisher (iafisher@protonmail.com)
Version: December 2018
"""
import re
import readline

from hera.utils import op_to_string, print_register_debug
from hera.vm import VirtualMachine


def debug(program):
    """Start the debug loop."""
    debugger = Debugger(program)
    debugger.loop()


_HELP_MSG = """\
Available commands:
    break <n>    Set a breakpoint on the n'th line of the program. When no
                 arguments are given, all current breakpoints are printed.

    help         Print this help message.

    next         Execute the current line.

    print <e>    Evaluate the expression and print its result. The expression
                 may be a register or a memory location, e.g. "M[123]".

    restart      Restart the execution of the program from the beginning.

    skip <n>     Skip to the n'th line of the program without executing any
                 of the lines in between.

    quit         Exit the debugger.

Command names may be abbreviated with a unique prefix, e.g. "n" for "next".
"""


class Debugger:
    """A class for debugging. External users should generally use the module-level
    `debug` function instead of this class.
    """

    def __init__(self, program):
        self.program = program
        # A map from instruction numbers (i.e., possible values of the program counter)
        # to human-readable line numbers.
        self.breakpoints = {}
        self.vm = VirtualMachine()

    def loop(self):
        self.print_current_op()

        while True:
            try:
                response = input(">>> ").strip()
            except (EOFError, KeyboardInterrupt):
                print()
                break

            if not response:
                continue

            if not self.handle_command(response):
                break

    def handle_command(self, response):
        """Parse the command and execute it. Return False if the loop should exit, and
        True otherwise.
        """
        cmd, *args = response.split()
        cmd = cmd.lower()
        if "break".startswith(cmd):
            self.exec_break(args)
        elif "continue".startswith(cmd):
            self.exec_continue(args)
        elif "next".startswith(cmd):
            self.exec_next(args)
        elif "print".startswith(cmd):
            self.exec_print(args)
        elif "restart".startswith(cmd):
            self.exec_restart(args)
        elif "skip".startswith(cmd):
            self.exec_skip(args)
        elif "help".startswith(cmd):
            print(_HELP_MSG)
        elif "quit".startswith(cmd):
            return False
        else:
            print("{} is not a known command.".format(cmd))

        return True

    def exec_break(self, args):
        if len(args) > 1:
            print("break takes zero or one arguments.")
            return

        if len(args) == 0:
            # TODO: In pdb, break with no arguments set a breakpoint at the current
            # line. Should I do that too?
            if self.breakpoints:
                for b in self.breakpoints.values():
                    print(b)
            else:
                print("No breakpoints set.")
        else:
            try:
                b = self.resolve_location(args[0])
            except ValueError as e:
                print("Error:", e)
            else:
                self.breakpoints[b] = self.get_breakpoint_name(b)

    def exec_next(self, args):
        if len(args) != 0:
            print("next takes no arguments.")
            return

        if self.vm.pc >= len(self.program):
            print("Program has finished executing. Press 'r' to restart.")
            return

        # The user expects that the next operation as it is written in the HERA file
        # will be executed. However, the next operation in the file may correspond to
        # multiple operations in the actual program, e.g. SET in the file becomes
        # SETLO and SETHI in the program. So we keep executing operations that
        # correspond to the same original operation (the `original` field on Op objects)
        # until we hit an operation with a different origin.
        original_op = self.program[self.vm.pc].original
        while (
            self.vm.pc < len(self.program)
            and self.program[self.vm.pc].original == original_op
        ):
            opc = self.vm.pc
            self.vm.exec_one(self.program[self.vm.pc])
            if opc == self.vm.pc:
                print("Program has finished executing.")
                return

        self.print_current_op()

    def exec_skip(self, args):
        if len(args) != 1:
            print("skip takes on argument.")
            return

        try:
            b = self.resolve_location(args[0])
        except ValueError as e:
            print("Error:", e)
        else:
            self.vm.pc = b
            self.print_current_op()

    def exec_continue(self, args):
        if len(args) != 0:
            print("continue takes no arguments.")
            return

        while self.vm.pc < len(self.program) and self.vm.pc not in self.breakpoints:
            opc = self.vm.pc
            self.vm.exec_one(self.program[self.vm.pc])
            if self.vm.pc == opc:
                print("Program has finished executing.")
                return

        self.print_current_op()

    # Match strings of the form "m[123]"
    _MEM_PATTERN = re.compile(r"[Mm]\[([0-9]+)\]")

    def exec_print(self, args):
        if len(args) != 1:
            print("print takes one argument.")
            return

        match = self._MEM_PATTERN.match(args[0])
        if match:
            index = int(match.group(1))
            print("M[{}] = {}".format(index, self.vm.access_memory(index)))
        elif args[0].lower() == "pc":
            print("PC = {}".format(self.vm.pc))
        else:
            try:
                v = self.vm.get_register(args[0])
            except ValueError:
                print("{} is not a valid register.".format(args[0]))
            else:
                print_register_debug(args[0], v, to_stderr=False)

    def exec_restart(self, args):
        if len(args) != 0:
            print("restart takes no arguments.")
            return

        self.vm.reset()
        self.print_current_op()

    def print_current_op(self):
        """Print the next operation to be executed. If the program has finished
        executed, nothing is printed.
        """
        if self.vm.pc < len(self.program):
            op = self.program[self.vm.pc].original or self.program[self.vm.pc]
            opstr = op_to_string(op)
            if op.name.location is not None:
                path = (
                    "<stdin>" if op.name.location.path == "-" else op.name.location.path
                )
                print("[{}, line {}]\n".format(path, op.name.location.line))
            print("{:0>4x}  {}".format(self.vm.pc, opstr))

    def resolve_location(self, b):
        """Resolve a user-supplied location string into an instruction number"""
        try:
            b = int(b)
        except ValueError:
            raise ValueError("could not parse argument.") from None

        for i, op in enumerate(self.program):
            if op.name.location.line == b:
                return i

        raise ValueError("could not find corresponding line.")

    def get_breakpoint_name(self, b):
        """Turn an instruction number into a human-readable location string with the
        file path and line number. More or less the inverse of `resolve_location`.
        """
        op = self.program[b].original or self.program[b]
        if op.name.location is not None:
            path = "<stdin>" if op.name.location.path == "-" else op.name.location.path
            return path + ":" + str(op.name.location.line)
        else:
            return str(op.name.location.line)



import pytest
from unittest.mock import patch

from hera.data import Op
from hera.utils import to_u16
from hera.vm import VirtualMachine


@pytest.fixture
def vm():
    return VirtualMachine()


def test_exec_one_delegates_to_setlo(vm):
    with patch("hera.vm.VirtualMachine.exec_setlo") as mock_exec_setlo:
        vm.exec_one(Op("SETLO", ["R1", 47]))
        assert mock_exec_setlo.call_count == 1
        assert mock_exec_setlo.call_args == (("R1", 47), {})


def test_setlo_with_positive(vm):
    vm.exec_setlo("R5", 23)
    assert vm.registers[5] == 23


def test_setlo_with_negative(vm):
    vm.exec_setlo("R9", -12)
    assert vm.registers[9] == to_u16(-12)


def test_setlo_with_max_positive(vm):
    vm.exec_setlo("R2", 127)
    assert vm.registers[2] == 127


def test_setlo_with_255(vm):
    vm.exec_setlo("R2", 255)
    assert vm.registers[2] == to_u16(-1)


def test_setlo_with_max_negative(vm):
    vm.exec_setlo("R2", -128)
    assert vm.registers[2] == to_u16(-128)


def test_setlo_clears_high_bits(vm):
    vm.registers[6] = 4765
    vm.exec_setlo("R6", 68)
    assert vm.registers[6] == 68


def test_setlo_increments_pc(vm):
    vm.exec_setlo("R9", -12)
    assert vm.pc == 1


def test_setlo_ignores_flags(vm):
    vm.flag_carry = True
    vm.flag_overflow = True
    vm.flag_sign = True
    vm.flag_zero = False
    vm.exec_setlo("R7", 0)
    assert vm.flag_carry
    assert vm.flag_overflow
    assert vm.flag_sign
    assert not vm.flag_zero


def test_setlo_does_not_set_zero_flag(vm):
    vm.exec_setlo("R7", 0)
    assert not vm.flag_zero


def test_setlo_does_not_set_sign_flag(vm):
    vm.exec_setlo("R7", -1)
    assert not vm.flag_sign


def test_setlo_does_not_change_R0(vm):
    vm.exec_setlo("R0", 20)
    assert vm.registers[0] == 0


def test_exec_one_delegates_to_sethi(vm):
    with patch("hera.vm.VirtualMachine.exec_sethi") as mock_exec_sethi:
        vm.exec_one(Op("SETHI", ["R1", 47]))
        assert mock_exec_sethi.call_count == 1
        assert mock_exec_sethi.call_args == (("R1", 47), {})


def test_sethi_with_positive(vm):
    vm.exec_sethi("R5", 23)
    assert vm.registers[5] == 5888


def test_sethi_with_max_positive(vm):
    vm.exec_sethi("R2", 255)
    assert vm.registers[2] == 65280


def test_sethi_does_not_clear_low_bits(vm):
    vm.registers[6] = 4765
    vm.exec_sethi("R6", 68)
    assert vm.registers[6] == 17565


def test_sethi_increments_pc(vm):
    vm.exec_sethi("R9", 12)
    assert vm.pc == 1


def test_sethi_ignores_flags(vm):
    vm.flag_carry = True
    vm.flag_overflow = True
    vm.flag_sign = True
    vm.flag_zero = False
    vm.exec_sethi("R7", 0)
    assert vm.flag_carry
    assert vm.flag_overflow
    assert vm.flag_sign
    assert not vm.flag_zero


def test_sethi_does_not_set_zero_flag(vm):
    vm.exec_sethi("R7", 0)
    assert not vm.flag_zero


def test_sethi_does_not_set_sign_flag(vm):
    vm.exec_sethi("R7", -1)
    assert not vm.flag_sign


def test_sethi_does_not_change_R0(vm):
    vm.exec_sethi("R0", 20)
    assert vm.registers[0] == 0


def test_exec_one_delegates_to_inc(vm):
    with patch("hera.vm.VirtualMachine.exec_inc") as mock_exec_inc:
        vm.exec_one(Op("INC", ["R1", 1]))
        assert mock_exec_inc.call_count == 1
        assert mock_exec_inc.call_args == (("R1", 1), {})


def test_inc_with_small_positive(vm):
    vm.exec_inc("R8", 6)
    assert vm.registers[8] == 6


def test_inc_with_max(vm):
    vm.exec_inc("R2", 32)
    assert vm.registers[2] == 32


def test_inc_with_previous_value(vm):
    vm.registers[5] = 4000
    vm.exec_inc("R5", 2)
    assert vm.registers[5] == 4002


def test_inc_with_previous_negative_value(vm):
    vm.registers[9] = to_u16(-12)
    vm.exec_inc("R9", 10)
    assert vm.registers[9] == to_u16(-2)


def test_inc_increments_pc(vm):
    vm.exec_inc("R1", 1)
    assert vm.pc == 1


def test_inc_sets_zero_flag(vm):
    vm.registers[7] = to_u16(-1)
    vm.exec_inc("R7", 1)
    assert vm.registers[7] == 0
    assert vm.flag_zero
    assert not vm.flag_sign


def test_inc_sets_sign_flag(vm):
    vm.registers[1] = 32765
    vm.exec_inc("R1", 5)
    assert vm.registers[1] == to_u16(-32766)
    assert not vm.flag_zero
    assert vm.flag_sign


def test_inc_sets_carry_flag(vm):
    vm.registers[8] = to_u16(-1)
    vm.exec_inc("R8", 1)
    assert vm.flag_carry
    assert not vm.flag_overflow


def test_inc_sets_overflow_flag(vm):
    vm.registers[8] = 32765
    vm.exec_inc("R8", 5)
    assert not vm.flag_carry
    assert vm.flag_overflow


def test_inc_ignores_incoming_carry(vm):
    vm.flag_carry = True
    vm.exec_inc("R8", 5)
    assert vm.registers[8] == 5
    assert not vm.flag_carry


def test_inc_does_not_affect_R0(vm):
    vm.exec_inc("R0", 1)
    assert vm.registers[0] == 0


def test_exec_one_delegates_to_dec(vm):
    with patch("hera.vm.VirtualMachine.exec_dec") as mock_exec_dec:
        vm.exec_one(Op("DEC", ["R1", 1]))
        assert mock_exec_dec.call_count == 1
        assert mock_exec_dec.call_args == (("R1", 1), {})


def test_dec_with_small_positive(vm):
    vm.exec_dec("R8", 6)
    assert vm.registers[8] == to_u16(-6)


def test_dec_with_max(vm):
    vm.exec_dec("R2", 32)
    assert vm.registers[2] == to_u16(-32)


def test_dec_with_previous_value(vm):
    vm.registers[5] = 4000
    vm.exec_dec("R5", 2)
    assert vm.registers[5] == 3998


def test_dec_with_previous_negative_value(vm):
    vm.registers[9] = to_u16(-12)
    vm.exec_dec("R9", 10)
    assert vm.registers[9] == to_u16(-22)


def test_dec_increments_pc(vm):
    vm.exec_dec("R1", 1)
    assert vm.pc == 1


def test_dec_sets_zero_flag(vm):
    vm.registers[7] = 1
    vm.exec_dec("R7", 1)
    assert vm.registers[7] == 0
    assert vm.flag_zero
    assert not vm.flag_sign


def test_dec_sets_sign_flag(vm):
    vm.registers[1] = 1
    vm.exec_dec("R1", 5)
    assert vm.registers[1] == to_u16(-4)
    assert not vm.flag_zero
    assert vm.flag_sign


def test_dec_sets_carry_flag(vm):
    vm.exec_dec("R8", 1)
    assert vm.flag_carry
    assert not vm.flag_overflow


def test_dec_sets_overflow_flag(vm):
    vm.registers[8] = to_u16(-32768)
    vm.exec_dec("R8", 5)
    assert vm.registers[8] == 32763
    assert not vm.flag_carry
    assert vm.flag_overflow


def test_dec_ignores_incoming_carry(vm):
    vm.flag_carry = True
    vm.registers[8] = 10
    vm.exec_dec("R8", 5)
    assert vm.registers[8] == 5
    assert not vm.flag_carry


def test_dec_does_not_affect_R0(vm):
    vm.exec_dec("R0", 1)
    assert vm.registers[0] == 0



"""Convenient interface for parsing, type-checking and optionally preprocessing a HERA
file.

Author:  Ian Fisher (iafisher@protonmail.com)
Version: December 2018
"""
import sys

from . import config, preprocessor
from .parser import parse_file
from .symtab import get_symtab
from .typechecker import typecheck


def load_program(path, *, preprocess=True):
    """Read the HERA program from the file at `path` and return the program represented
    as a list of Op objects.

    This function combines parsing, type-checking and preprocessing (if `preprocess` is
    set to True).
    """
    config.ERROR_COUNT = config.WARNING_COUNT = 0

    program = parse_file(path, expand_includes=True, allow_stdin=True)

    # Print a newline if the program came from standard input, so that the
    # program and its output are visually separate.
    if path == "-":
        print()

    symtab = get_symtab(program)

    typecheck(program, symtab)
    if config.ERROR_COUNT > 0:
        sys.exit(3)

    if preprocess:
        program = preprocessor.preprocess(program, symtab)

    return program



import pytest

from hera.symtab import HERA_DATA_START
from hera.vm import VirtualMachine


@pytest.fixture
def vm():
    return VirtualMachine()


def test_integer_fills_in_memory(vm):
    vm.exec_integer(42)
    assert vm.memory[HERA_DATA_START] == 42


def test_integer_increments_data_counter(vm):
    vm.exec_integer(42)
    assert vm.dc == HERA_DATA_START + 1


def test_integer_increments_pc(vm):
    vm.exec_integer(42)
    assert vm.pc == 1


def test_dskip_increments_data_counter(vm):
    vm.exec_dskip(10)
    assert vm.dc == HERA_DATA_START + 10


def test_dskip_increments_pc(vm):
    vm.exec_dskip(10)
    assert vm.pc == 1


def test_lp_string_fills_in_memory(vm):
    vm.exec_lp_string("hello")
    assert vm.memory[HERA_DATA_START] == 5
    assert vm.memory[HERA_DATA_START + 1] == ord("h")
    assert vm.memory[HERA_DATA_START + 2] == ord("e")
    assert vm.memory[HERA_DATA_START + 3] == ord("l")
    assert vm.memory[HERA_DATA_START + 4] == ord("l")
    assert vm.memory[HERA_DATA_START + 5] == ord("o")


def test_lp_string_increments_data_counter(vm):
    vm.exec_lp_string("hello")
    assert vm.dc == HERA_DATA_START + 6


def test_lp_string_increments_pc(vm):
    vm.exec_lp_string("hello")
    assert vm.pc == 1



"""Preprocess HERA programs to convert pseudo-instructions and resolve labels.

Author:  Ian Fisher (iafisher@protonmail.com)
Version: December 2018
"""
from .data import Op, Token
from .utils import is_symbol, REGISTER_BRANCHES, to_u16


def preprocess(program, symtab):
    """Preprocess the program (a list of Op objects) into valid input for the
    exec_many method on the VirtualMachine class.

    This function does the following
        - Replaces pseudo-instructions with real ones.
        - Resolves labels into their line numbers.
    """
    program = [substitute_label(op, symtab) for op in program]
    program = [
        op._replace(original=old_op) for old_op in program for op in convert(old_op)
    ]
    program = [op for op in program if op.name not in ("LABEL", "DLABEL", "CONSTANT")]
    return program


def substitute_label(op, symtab):
    """Substitute any label in the instruction with its concrete value."""
    for i, arg in enumerate(op.args):
        if isinstance(arg, Token) and arg.type == "SYMBOL":
            op.args[i] = symtab[arg]
    return op


def convert(op):
    """Convert a pseudo-instruction into a list of real instructions."""
    if op.name in REGISTER_BRANCHES and isinstance(op.args[0], int):
        lbl = op.args[0]
        new_ops = [
            Op("SETLO", ["R11", lbl & 0xFF]),
            Op("SETHI", ["R11", lbl >> 8]),
            Op(op.name, ["R11"]),
        ]
    elif op.name in REGISTER_BRANCHES and is_symbol(op.args[0]):
        # This clause is only necessary for the symbol table generator--see the note in
        # `convert_call` below.
        lbl = op.args[0]
        new_ops = [
            Op("SETLO", ["R11", lbl]),
            Op("SETHI", ["R11", lbl]),
            Op(op.name, ["R11"]),
        ]
    elif op.name == "SET":
        new_ops = convert_set(*op.args)
    elif op.name == "CMP":
        new_ops = [Op("FON", [8]), Op("SUB", ["R0", op.args[0], op.args[1]])]
    elif op.name == "CON":
        new_ops = [Op("FON", [8])]
    elif op.name == "COFF":
        new_ops = [Op("FOFF", [8])]
    elif op.name == "CBON":
        new_ops = [Op("FON", [16])]
    elif op.name == "CCBOFF":
        new_ops = [Op("FOFF", [24])]
    elif op.name == "MOVE":
        new_ops = [Op("OR", [op.args[0], op.args[1], "R0"])]
    elif op.name == "SETRF":
        new_ops = convert_set(*op.args) + convert(Op("FLAGS", [op.args[0]]))
    elif op.name == "FLAGS":
        new_ops = [Op("FOFF", [8]), Op("ADD", ["R0", op.args[0], "R0"])]
    elif op.name == "HALT":
        new_ops = [Op("BRR", [0])]
    elif op.name == "NOP":
        new_ops = [Op("BRR", [1])]
    elif op.name == "CALL":
        new_ops = convert_call(*op.args)
    elif op.name == "NEG":
        new_ops = [Op("FON", [8]), Op("SUB", [op.args[0], "R0", op.args[1]])]
    elif op.name == "NOT":
        new_ops = [
            Op("SETLO", ["R11", 0xFF]),
            Op("SETHI", ["R11", 0xFF]),
            Op("XOR", [op.args[0], "R11", op.args[1]]),
        ]
    else:
        new_ops = [op]
    # Copy over the line and column information from the original token.
    return [Op(copy_token(new_op.name, op.name), new_op.args) for new_op in new_ops]


def convert_set(d, v):
    if isinstance(v, int):
        v = to_u16(v)
        lo = v & 0xFF
        hi = v >> 8
        return [Op("SETLO", [d, lo]), Op("SETHI", [d, hi])]
    else:
        return [Op("SETLO", [d, v]), Op("SETHI", [d, v])]


def convert_call(a, l):
    if isinstance(l, int):
        return [
            Op("SETLO", ["R13", l & 0xFF]),
            Op("SETHI", ["R13", l >> 8]),
            Op("CALL", [a, "R13"]),
        ]
    elif is_symbol(l):
        # This clause is only necessary for when the symbol table generator calls
        # `convert` to calculate the values of labels in the final program. At that
        # point, labels have not yet been resolved, but we want to make sure we generate
        # the same number of instructions so that the label offsets are correct.
        return [
            Op("SETLO", ["R13", l]),
            Op("SETHI", ["R13", l]),
            Op("CALL", [a, "R13"]),
        ]
    else:
        return [Op("CALL", [a, l])]


def copy_token(val, otkn):
    """Convert the string `val` into a Token with the same line and column numbers as
    `otkn`.
    """
    if isinstance(otkn, Token):
        return Token(otkn.type, val, loc=otkn.location)
    else:
        return val
